#ifndef HERMES_SHM_TEST_UNIT_SOLVERS_TEST_INIT_H_
#define HERMES_SHM_TEST_UNIT_SOLVERS_TEST_INIT_H_

#include "basic_test.h"
#include "hermes_shm/solver/nonlinear_least_squares.h"

#endif  // HERMES_SHM_TEST_UNIT_SOLVERS_TEST_INIT_H_