#ifndef MOD_NAME_AUTOGEN_METHODS_H_
#define MOD_NAME_AUTOGEN_METHODS_H_

#include <chimaera/chimaera.h>

/**
 * Auto-generated method definitions for MOD_NAME
 */

namespace chimaera::MOD_NAME {

namespace Method {
// Inherited methods
GLOBAL_CONST chi::u32 kCreate = 0;
GLOBAL_CONST chi::u32 kDestroy = 1;

// MOD_NAME-specific methods
GLOBAL_CONST chi::u32 kCustom = 10;
GLOBAL_CONST chi::u32 kCoMutexTest = 20;
GLOBAL_CONST chi::u32 kCoRwLockTest = 21;
GLOBAL_CONST chi::u32 kWaitTest = 23;
}  // namespace Method

}  // namespace chimaera::MOD_NAME

#endif  // MOD_NAME_AUTOGEN_METHODS_H_
